;---------------------------------------------------------------------
; Programming Conventions:
; Strict:
; - Core revision 6 instruction set
; - R7 is the Stack Pointer	(software)
; - R6 holds the return address for subroutines (JAL)
; - Full-descending Stack
; Loose guidelines:
; - R0, R1, R2 are used for arguments / results
;              Values may change on subroutine calls
; - R3, R4, R5 are used for local variables
;              Values are preserved on subroutine calls
;---------------------------------------------------------------------
;---------------------------------------------------------------------
;----------- I/O MAP  -----------

IOBASE=		0x20
IRQEN=		0x20
PFLAGS=		0x21	; read only
PALETTE=	0x21	; write only
UARTDAT=	0x22
OUTPORT=	0x23
TIMER=		0x24
CLRHDE=		0x26	; Clear IRQ strobe on write. Data ignored
CLRVDE=		0x27	; Clear IRQ strobe on write. Data ignored

;--------- Interrupt enable --------
IRQRXEN=	1
IRQTXEN=	2
IRQTIMEN=	4
IRQHDEEN=	8
IRQVDEEN=	16
;--------- PFLAGS register bits -----------
UARTDV=		1		; Data valid (UART RX)
UARTRDY=	2		; Transmitter ready (UART TX)
UARTOVER=	4		; RX overrun Error
UARTFER=	8		; RX Frame Error
HDE=		16		; Horizontal display enable
VDE=		32		; Vertical display enable
TIMOV=		0x8000	; Timer overflow (sign bit)

;--------------------------------------
;------- HEADER for bootloader --------
;--------------------------------------

		org		0x40-4
		
		word	0x4CFF			; mark
		word	pinit			; destination address
		word	pend-pinit		; size (words)
		word	pstart			; execution address

;--------------------------------------
;------------- CODE -----------------
;--------------------------------------
pinit:

;------------------------------------------------------------------------
; fast address variables (single LDI for address loading)
;------------------------------------------------------------------------
		org		0x40
bvar:
nint:	word	0


;-------------------------------------------------------------------
; I/O subroutines
;-------------------------------------------------------------------

;------------------------------------------------------------
; Prints ASCIIZ (16-bit packed, little endian) string on UART 
; parameters:
;	R0: pointer to string
;	R6: return address (JAL)
; returs:
; 	R0-R2: modified

putsle:	ldi		r2,UARTDAT
putsle1:
		ld		r1,(r0)
		andi	r1,0xff
		jz		putsle3
		st		(r2),r1
		
		ld		r1,(r0)
		rori	r1,r1,8
		andi	r1,0xff
		jz		putsle3
		st		(r2),r1
		addi	r0,1
		jr		putsle1

putsle3:
		jind	r6
;----------------------------------------------
; Sends character to UART
; parameters:
;	R0: character to send
;	R6: return address (JAL)
; returns:
; 	R1: modiffied
; notes: CPU clock stops if UART not ready for transmission

putch:	ldi		r1,UARTDAT
		st		(r1),r0		
		jind	r6			

;----------------------------------------------
; Receives character from UART
; parameters:
;	R6: return address (JAL)
; returns:
;	R0: reveived char
; 	R1: modiffied
; notes: CPU clock stops until a character is received

 getch:	ldi		r1,UARTDAT	
 		ld		r0,(r1)		
 		jind	r6			

;----------------------------------------------
; Prints decimal number
; parameters:
;	R0: number tyo print
;	R6: return address (JAL)
 
; returns:
;	R0, R1: modiffied
; retorna: R0 y R1 modificados
; notes:
;	Avoids printing zeroes on the left
;   Stack used for temporary digit storage

prtdec:
		subi	r7,1
		st		(r7),r6

		ldi		r1,0		; end of string mark
		subi	r7,1		; to stack
		st		(r7),r1
prtd0:	ldi		r1,0		; R0=R0/10, R1=remainder
		ldi		r6,16		
prtd1:	add		r0,r0,r0
		adc		r1,r1,r1
		cmpi	r1,10
		jnc		prtd2
		subi	r1,10
		addi	r0,1
prtd2:	subi	r6,1
		jnz		prtd1
		ldi		r6,'0'		; remainder in ASCII goes to stack
		add		r1,r1,r6
		subi	r7,1
		st		(r7),r1
		or		r0,r0,r0	; repeat until zero
		jnz		prtd0

prtd3:	ld		r0,(r7)		; retrieve character from stack
		jz		prtd9		; end of string?
		addi	r7,1
		jal		putch
		jr		prtd3

prtd9:	addi	r7,1
		ld		r6,(r7)
		addi	r7,1
		jind	r6

;--------------------------------------------
;  VIDEO functions
;--------------------------------------------
; Clear Screen
; R0 : filling pixel: values 0 to 3
; no reg modiffied
cls:	subi	r7,2
		st		(r7),r0
		st		(r7+1),r6
		andi	r0,3
		ldpc	r6
		word	pxtab
		add		r6,r6,r0
		ld		r0,(r6)
		ldpc	r6
		word	0x6A00
bcls:	st		(r6),r0
		st		(r6+1),r0
		st		(r6+2),r0
		st		(r6+3),r0
		addi	r6,4
		jnz		bcls
		
		ld		r0,(r7)
		ld		r6,(r7+1)
		addi	r7,2
		jind	r6
pxtab:	word	0x0000
		word	0x00FF
		word	0xFF00
		word	0xFFFF
		
;---------------------------------------------------------
; Text character drawing (8x16 font => 80 x 30 characters)
;  pixel values are 0 for background or 3 for foreground
;    R0: ASCII code
;    R1: X pos (0 to 79)
;    R2: Y pos (0 to 223)
;    no reg modiffied
vputxy:	subi	r7,5
		st		(r7),r0
		st		(r7+1),r1
		st		(r7+2),r2
		st		(r7+3),r3
		st		(r7+4),r6
		rori	r6,r2,16-6	; Y*64
		rori	r2,r2,16-4	; Y*16
		add		r2,r2,r6	; Y*80
		add		r2,r2,r1	; Y*80 + X
		ldpc	r1
		word	0x6A00
		add		r2,r2,r1	; r2=video[Y*80+X]
		subi	r0,32		; if (chr<32 || chr>=96) chr=32
		jmi		vpxy0
		cmpi	r0,96
		jnc		vpxy1
vpxy0:	ldi		r0,0
vpxy1:	rori	r0,r0,16-3	; CHR*8
		ldpc	r1
		word	font
		add		r1,r1,r0	; character pixels
		ldi		r3,8
		
vpxy2:	ld		r0,(r1)
		andi	r0,255
		rori	r6,r0,8
		or		r0,r0,r6
		st		(r2),r0
		addi	r2,80
		ld		r0,(r1)
		rori	r0,r0,8
		andi	r0,255
		rori	r6,r0,8
		or		r0,r0,r6
		st		(r2),r0
		addi	r2,80
		addi	r1,1
		subi	r3,1
		jnz		vpxy2
				
		ld		r0,(r7)
		ld		r1,(r7+1)
		ld		r2,(r7+2)
		ld		r3,(r7+3)
		ld		r6,(r7+4)
		addi	r7,5
		jind	r6

;-------------------------------------------------------------------
; INTERRUPTS
;-------------------------------------------------------------------
		org	0x100
irq0:	;jr	RXIRQ
		org	0x104
irq1:	;jr	TXIRQ
		org	0x108
irq2:	;jr	timerIRQ
		org	0x10c
irq3:	;jr	myIRQ ;hdeIRQ
		org	0x110
irq4:	;jr	vdeIRQ

myIRQ:
		subi	r7,3
		st		(r7),r0
		st		(r7+1),r1
		st		(r7+2),r2

		ldi		r1,IOBASE		
		st		(r1+CLRVDE-IOBASE),r1	; Clear IRQ

		ldi		r2,nint
		ld		r0,(r2)
		addi	r0,1
		st		(r2),r0
		andi	r0,3
		jnz		iend

		ld		r0,(r1+OUTPORT-IOBASE)
		add		r0,r0,r0
		rori	r2,r0,3
		andi	r2,1
		xori	r2,1
		or		r0,r0,r2
		st		(r1+OUTPORT-IOBASE),r0
		
iend:	ld		r0,(r7)
		ld		r1,(r7+1)
		ld		r2,(r7+2)
		addi	r7,3
		reti


;-------------------------------------------------------------------
;
;		MAIN
;
;-------------------------------------------------------------------
pstart:	
start:	
		ldpc	r7
		word	0x6A00			; Stack below Video
	
		ldpc	r0
		word	txt
		jal		putsle

		ldi		r0,0
		jal		cls
		ldi		r0,0
		ldi		r1,24
		ldi		r2,240-32
bu0:	jal		vputxy
		addi	r0,1
		cmpi	r0,255
		jz		bu1
		addi	r1,1
		cmpi	r1,32+24
		jnz		bu0
		ldi		r1,24
		addi	r2,16
		jr		bu0	
		
bu1:
		ldi		r5,IOBASE	
		ldi		r0,IRQVDEEN					; Enable VDE IRQ
		st		(r5+IRQEN-IOBASE),r0
		
mbuc:	;ld		r0,(r5+TIMER-IOBASE)
		ldi		r0,nint
		ld		r0,(r0)
		jal		prtdec
		ldi		r0,'\n'
		jal		putch
mdl1:	ld		r0,(r5+PFLAGS-IOBASE)
		andi	r0,VDE
		jnz		mdl1
mdl2:	ld		r0,(r5+PFLAGS-IOBASE)
		andi	r0,VDE
		jz		mdl2

		jr 		mbuc

txt:	asczle "\nPruebas Video\n"
		
font:	; Chars 32 to 127, two lines per word (LSB first)
		include "font8x16.inc"
pend:
;-------------------------------------------------------------------
; Variables
;-------------------------------------------------------------------

	


