//--------------------------------------------------------------------
// Sistema con CPU, RAM y UART
//--------------------------------------------------------------------

`include "cpuV6.v"
`include "uart_simple.v"
`include "VGA.v"

module SYSTEM (
	input clk,		
	input reset,
	
	// XRAM
	inout  [15:0]xd,
	output [15:0]xa,
	output xweb,
	output xoeb,

	// UART
	output txd,
	input rxd,
	// LEDs
	output [7:0]led,
	
	// VGA
	output hsyn,
	output vsyn,
	output [3:0]video
);

parameter FCLK=24000000;
parameter BAUD=115200;

//////////////////////// BUSSES ////////////////////////
wire [15:0]cdi;	// CPU Data bus, Input
wire [15:0]cdo;	// CPU Data bus, Output
wire [15:0]ca;	// CPU Address bus
wire we;		// Write Enable

/////////////////////////////// MEMORY ////////////////////////////////
// Internal boot ROM (32 words), Combinational

wire iromcs = (ca[15:5]==0);

reg [15:0]irom[0:31];
wire [15:0]romdo = irom[ca[4:0]];

initial begin
`ifdef SIM
  $readmemh("rom_prueba.hex", irom);
`else
  $readmemh("rom_boot.hex", irom);
`endif
end

	
////////////////////////////////////////////////
// External SRAM

wire xwe = we & (~clk) & (~vrd);
wire xoe = (~we) | vrd;

assign xa = vrd ? va : ca;
assign xd = (xoe) ? 16'bz : cdo;

assign xoeb=~xoe;
assign xweb=~xwe;


//////////////////////////////////////////////////
//////					CPU					//////
//////////////////////////////////////////////////

assign cdi = iocs ? iodo : (iromcs ? romdo : xd);

wire stopcpu; // stop CPU if 1
wire cclk = clk | stopcpu;

`ifdef SIM
CPU_GUSV6 #(.VECTORBASE(16'h0000)) GUScpu0( 
`else
CPU_GUSV6 #(.VECTORBASE(16'h0100)) GUScpu0( 
`endif
	.clk(cclk), .we(we), .ca(ca), .cdo(cdo), .cdi(cdi), 
	.reset(reset), .irq(irq), .ivector(ivector)
);


// IRQEN reg
reg [4:0]irqen=0;	
always @(posedge clk or posedge reset ) 
	if (reset) irqen<=0;
	else irqen <= iowe0 ? cdo[4:0] : irqen;
// IRQS
wire [4:0]irqs={ virq & irqen[4], hirq & irqen[3],
	tflag& irqen[2],uflags[1] & irqen[1],uflags[0] & irqen[0]};
wire irq = |irqs;
wire [2:0]ivector = irqs[0] ? 3'd0 : (irqs[1]? 3'd1 : (irqs[2]? 3'd2 : (irqs[3]? 3'd3 : 3'd4)));

/////////////////////////////////////////////////////////
///////////////////// Peripherals ///////////////////////
/////////////////////////////////////////////////////////

wire iocs;		// IO select ($0020 to $003F)

assign iocs=(ca[15:5]==11'b00000000001);
// Write strobes
wire iowe0=(we)&(iocs)&(ca[2:0]==0);
wire iowe1=(we)&(iocs)&(ca[2:0]==1);
wire iowe2=(we)&(iocs)&(ca[2:0]==2);
wire iowe3=(we)&(iocs)&(ca[2:0]==3);
wire iowe4=(~stopcpu)&(we)&(iocs)&(ca[2:0]==4);
//wire iowe5=(we)&(iocs)&(ca[2:0]==5);
wire iowe6=(we)&(iocs)&(ca[2:0]==6);
wire iowe7=(we)&(iocs)&(ca[2:0]==7);
// Read strobes
wire iore2=(~we)&(iocs)&(ca[2:0]==2);
wire iore4=(~stopcpu)&(~we)&(iocs)&(ca[2:0]==4);

reg [7:0]outport=0;
always @(posedge cclk) if (iowe3) outport<=cdo[7:0];
reg [15:0]palette=16'b1111_0100_0010_0001;
always @(posedge cclk) if (iowe1) palette<=cdo;
// Timer
// prescaler for 1MHz timer clk
localparam PRESC=FCLK/1000000;
localparam NBPRE=$clog2(PRESC);
reg [NBPRE-1:0]presc=0;
wire prestc = &(presc | (~(PRESC-1)) );

reg [15:0]tmax=16'hffff;
reg [15:0]timer=0;
reg tflag=0;
wire timetc = (timer==tmax);
always @(posedge clk) begin
	presc <= (iowe4 |prestc) ? 0 : presc+1;
	timer <= (iowe4 | (timetc & prestc)) ? 0 : timer + prestc;
	tmax  <=  iowe4 ? cdo : tmax;
	tflag <= (timetc & prestc) ? 1 : ((iowe4 | iore4)? 0 : tflag);
end

/////////////////////////////
// I/O data to CPU MUX
reg [15:0]iodo;
always@*
  case (ca[2:0])
     0 : iodo <=  {15'h0,irqen}; // IRQ Enable
     1 : iodo <=  {tflag,9'h0,vde,hde,uflags}; // PFLAGS
     2 : iodo <=  {8'h0,urxdata}; // UART data
     3 : iodo <=  {8'h0,outport}; // output port
     4 : iodo <=  timer;
     default : iodo <= 16'hxxxx;
  endcase

assign stopcpu = (iore2 & (~uflags[0])) | (iowe2&(~uflags[1])) | vrd;

/////////////////////////////
// UART
localparam DIVIDER=(FCLK+(BAUD/2))/BAUD; 

wire [7:0]urxdata;
wire [3:0]uflags;

UART_core #(.DIVIDER(DIVIDER)) uart0 
	( .clk(clk), .d(cdo[7:0]), .wr(iowe2&(~stopcpu)), .rd(iore2&(~stopcpu)), .q(urxdata),
		.rxvalid(uflags[0]), .txrdy(uflags[1]),
		.rxoverr(uflags[2]), .rxframeer(uflags[3]),
		.nstop(1'b0),
		.txd(txd), .rxd(rxd)
	);
	
/////////////////////////////
// VIDEO
wire [15:0]va;
wire vrd;

videoVGA video0
( .clk(clk), .hsyn(hsyn), .vsyn(vsyn), .video(video), .hde(hde), .vde(vde),
	.vrd(vrd), .va(va), .d(xd), .palette (palette)
	
);
wire hde, vde;
reg [1:0]ohde=0; reg [1:0]ovde=0;
reg virq=0; reg hirq=0;
always @(posedge cclk) begin
	ohde<={ohde[0],hde};
	ovde<={ovde[0],vde};
	hirq<= (ohde[1] & (~ohde[0])) ? 1 : (iowe6 ? 0 : hirq);
	virq<= (ovde[1] & (~ovde[0])) ? 1 : (iowe7 ? 0 : virq);
end

assign led=outport;

endmodule


